"""Evolutionary informatics package."""

from .lib import tobin
from .lib import tofetch
from .lib import tofile
from .lib import toinsdc
from .lib import tolog
from .lib import totax
from .lib import version
from .tolkein import cli
